# -*- coding: utf-8 -*-
__all__ = ('APPLICATION_COMMANDS', 'APPLICATIONS', 'CHANNELS', 'CLIENTS', 'DISCOVERY_CATEGORIES', 'EMOJIS', 'EULAS',
    'GUILDS', 'INTEGRATIONS', 'INVITES', 'KOKORO', 'MESSAGES', 'ROLES', 'TEAMS', 'USERS', 'start_clients',
    'stop_clients', )

import sys, gc
from time import perf_counter
from threading import current_thread

from ..backend.utils import WeakValueDictionary
from ..backend.futures import Future, sleep, CancelledError, future_or_timeout, Task
from ..backend.event_loop import EventThread

Client = NotImplemented

if (sys.hash_info.width >= 64):
    #if we have 64 bit system we can use array instead of list
    from array import array as Array
    def create_array():
        return Array('Q')
else:
    create_array = list
    
if sys.implementation.name == 'cpython':
    #on CPython bisect is 4~ times faster.
    import bisect
    _relative_index = bisect.bisect_left
    del bisect

else:
    def _relative_index(array, value):
        """
        Return the index where to insert item `value`  in list `array`, assuming `array` is sorted.
        
        Parameters
        ----------
        array : `Any`
            Any sorted indexable type.
        value : `Any`
            The object, which index in the array will be returned.
        
        Returns
        -------
        index : `int`
        """
        bot = 0
        top = len(array)
        while True:
            if bot < top:
                half = (bot+top)>>1
                if array[half] < value:
                    bot = half+1
                else:
                    top = half
                continue
            break
        return bot

class ClientDictionary(object):
    """
    A dictionary like data object for storing directly ``Client`` objects.
    
    Attributes
    ----------
    _elements : `list` of ``Client``
        The stored ``Client`` objects sorted by their id.
    _ids : (`array` of `uint64`) if sys.bit >= 64 else (`list` of `int`)
        The client's from `._elements` in a sorted form.
    _next : `int`
        The next added ``Client``'s auto id, if it has no id set.
    """
    __slots__ = ('_elements', '_ids', '_next',)
    
    def __init__(self):
        """
        Creates a ``ClientDictionary``.
        """
        self._elements = []
        self._ids = create_array()
        self._next = 1
    
    def append(self, client):
        """
        Appends the container with the passed ``Client``. If the client has no `.id` set yet, auto generates one for
        it.
        
        Parameters
        ----------
        client : ``Client``
            The client to add to the container.
        
        Raises
        ------
        RuntimeError:
            If an object with it's respective `.id` same as `client`'s is already added to the container.
        """
        id_ = client.id
        
        if id_ < 4194304:
            id_ = self._next
            client.id = id_
            self._next = id_+1
        
        array = self._ids
        index = _relative_index(array, id_)
        
        if index == len(array): #put it at the last place
            array.append(id_)
            self._elements.append(client)
            return
        
        if array[index] == id_:  #this object is already at the container, lets check it!
            if self._elements[index] is client:
                return
            #hah, got u!
            raise RuntimeError('Two different objects added with same id')
        
        #insert it at the right place
        array.insert(index, id_)
        self._elements.insert(index, client)
    
    def remove(self, client):
        """
        Removes the passed ``Client`` from the container. If the ``Client`` is not at the container, then will not
        raise.
        
        Parameters
        ----------
        client : ``Client``
        """
        id_ = client.id
        
        if id_ == 0:
            return #not in the container
        
        array = self._ids
        index = _relative_index(array, id_)
        
        if index == len(array): #not in the container
            return
        
        if array[index] != id_:  #this object is not at the container, lets remove it
            return
        
        del array[index]
        del self._elements[index]
    
    #familiar to normal remove
    def remove_by_id(self, id_):
        """
        Removes a ``Client`` by it's `.id` from the container. If no client is with the specified `.id` at the
        container, will not raise.
        
        Parameters
        ----------
        id_ : `int`
            The `.id` of a client to remove.
        """
        array = self._ids
        index = _relative_index(array, id_)
        
        if index == len(array): #not in the container
            return
        
        if array[index] != id_:  #this object is not at the container, lets remove it
            return
        
        del array[index]
        del self._elements[index]
    
    def update(self, client, new_id):
        """
        Updates a client at the container with a new id. Sets the new `.id` of the client and modifies it's position
        at the container if needed.
        
        Parameters
        ----------
        client : ``Client``
            The client, what's `.id` will be updated.
        new_id : `int`
            The new `.id` of the client
            
        Raises
        ------
        RuntimeError
            If an object with it's respective `.id` is same as `new_id` is already added to the container.
        """
        old_id = client.id
        client.id = new_id
        array = self._ids
        old_index = _relative_index(array, old_id)
        
        if old_index == len(array) or array[old_index] != old_id: #not in the container?
            self.append(client)
            return
        
        new_index=_relative_index(array,new_id)
        
        #above or under?
        if old_index < new_index:
            move = 1
            new_index-=1
        elif old_index > new_index:
            move = -1
        else:
            if array[new_index] == new_id:
                #hah, got u!
                raise RuntimeError('Two different objects added with same id')
            return

        #move ids
        index = old_index
        while True:
            if index == new_index:
                break
            array[index] = array[index+move]
            index += move
        #put our at the right place
        array[index] = new_id
        
        elements = self._elements
        #move object
        index = old_index
        while True:
            if index == new_index:
                break
            elements[index] = elements[index+move]
            index += move
        #put our at the right place
        elements[index] = client

    def __getitem__(self, id_):
        """Returns the client for the specified id."""
        array = self._ids
        index = _relative_index(array, id_)
        if index == len(array) or array[index] != id_:
            raise KeyError(f'{id_!r} is not in the {self.__class__.__name__}')
        return self._elements[index]
    
    def get(self, id_, default=None):
        """
        Returns the client of the client dictionary for the given id.
        
        Parameters
        ----------
        default : `Any`, Optional
            Default value to return value if the client is not found.
        
        Returns
        -------
        client : ``Client`` or `default`
        """
        array = self._ids
        index = _relative_index(array, id_)
        if index == len(array) or array[index] != id_:
            return default
        return self._elements[index]
    
    def __contains__(self, client):
        """Returns whether the client is at the container."""
        id_ = client.id
        array = self._ids
        index = _relative_index(array, id_)
        if index == len(array):
            return False
        return (array[index] == id_)
    
    def __len__(self):
        """Returns the length of the container."""
        return len(self._ids)
    
    def index(self, client):
        """
        Returns the index, where the container stores the specific client.
        
        Parameters
        ----------
        client : ``Client``
            The client, what's index will be returned.
        
        Returns
        -------
        index : `int`
        
        Raises
        ------
        ValueError
            The client is not at the container.
        """
        id_ = client.id
        array = self._ids
        index = _relative_index(array, id_)
        if index == len(array) or array[index] != id_:
            raise ValueError(f'{client!r} is not in the {self.__class__.__name__}')
        return index
        
    def __iter__(self):
        """Returns an iterator over the clients stored by the container."""
        return iter(self._elements)

    def __reversed__(self):
        """Returns a reversed iterator over the clients stored by the container."""
        return reversed(self._elements)
    
    def count(self, client):
        """
        Counts how much time the passed client is at the container. Because the container contains only unique
        elements, it can return either `0` or `1`.
        
        Parameters
        ----------
        client : ``Client``
            The client to count.
        
        Returns
        -------
        count : `int`
        """
        id_ = client.id
        array = self._ids
        index = _relative_index(array, id_)
        if index == len(array):
            return 0
        if array[index] == id_:
            return 1
        return 0
    
    def copy(self):
        """
        Copies the container.
        
        Returns
        -------
        new : ``ClientDictionary``
        """
        new = list.__new__(type(self))
        new._ids = self._ids.copy()
        new._elements = self._elements.copy()
        new._next = self._next
        return new
    
    def __repr__(self):
        """Returns the representation of the container."""
        result = [self.__class__.__name__,'([']
        elements = self._elements
        stop = len(elements)
        if stop:
            stop -=1
            index = 0
            while index < stop:
                result.append(repr(elements[index]))
                result.append(', ')
                index += 1
            result.append(repr(elements[index]))
        result.append('])')
        
        return ''.join(result)


CHANNELS = WeakValueDictionary()
CLIENTS = ClientDictionary()
EMOJIS = WeakValueDictionary()
GUILDS = WeakValueDictionary()
INTEGRATIONS = WeakValueDictionary()
MESSAGES = WeakValueDictionary()
ROLES = WeakValueDictionary()
TEAMS = WeakValueDictionary()
USERS = WeakValueDictionary()
DISCOVERY_CATEGORIES = WeakValueDictionary()
EULAS = WeakValueDictionary()
APPLICATIONS = WeakValueDictionary()
INVITES = WeakValueDictionary()
APPLICATION_COMMANDS = WeakValueDictionary()

def start_clients():
    """
    Starts up all the not running clients.
    
    Can be called from any thread.
    """
    for client in CLIENTS:
        if client.running:
            continue
        
        Task(client.connect(), KOKORO)
    
    if (current_thread() is not KOKORO):
        KOKORO.wake_up()

def stop_clients():
    """
    Stops all the running clients.
    
    Can be called from any thread.
    """
    for client in CLIENTS:
        if client.running:
            Task(client.disconnect(), KOKORO)
    
    if (current_thread() is not KOKORO):
        KOKORO.wake_up()

KOKORO = EventThread(daemon=False, name='KOKORO')

GC_CYCLER = KOKORO.cycle(1200.)

if sys.implementation.name == 'pypy':
    def manual_gc_call(cycler):
        gc.collect()
    
    GC_CYCLER.append(manual_gc_call, (1<<31)-1)
    
    del manual_gc_call

HEARTBEAT_TIMEOUT = 20.0

class Kokoro(object):
    """
    A beater for Discord gateways.
    
    Attributes
    ----------
    beat_task : `None` or ``Task``
        The beat task of kokoro's gateway. Set by ``._task`` meanwhile it's gateway is connected.
    beat_waiter : `None` or ``Future``
        `.beater` task's waiter. Sleeps `.beater` between two beat.
    beater : `None` or ``Task`` of ``._keep_beating``
        A task what sends a beating payload to Discord. Set only meanwhile kokoro is in the middle of a beat.
    gateway : ``DiscordGateway`` or ``DiscordGatewayVoice``
        Kokoro's owner gateway.
    interval : `float`
        The time needed to be waited between two beats.
    last_answer : `float`
        The time of the last answered heartbeat event in `perf_counter` time.
    last_send : `float`
        The last time when kokoro sent a heartbeat event in `last_send` time.
    latency : `float`
        The time interval between the last beat and the response's time. Defaults to `.DEFAULT_LATENCY` if no beat was
        done yet.
    running : `bool`
        Whether the kokoro should run and restart itself.
    should_beat : `bool`
        Whether the kokoro should keep beating.
    task : `None` or ``Task`` of ``._start``
        The main keep alive task of kokoro.
    ws_waiter : `None` or ``Future``
        The waiter of kokoro, what waits for it's gateway to connect it's websocket.
    
    Class attributes
    ----------------
    DEFAULT_LATENCY : `float` = `9999.0`
        The default latency of kokoro. Better than using `inf`.
    """
    __slots__ = ('beat_task', 'beat_waiter', 'beater', 'gateway', 'interval', 'last_answer', 'last_send', 'latency',
        'running', 'should_beat', 'task', 'ws_waiter')
    
    DEFAULT_LATENCY = 9999.0
    
    async def __new__(cls, gateway):
        """
        Creates a ``Kokoro`` instance bound to it's gateway.
        
        This method is a coroutine.
        
        Parameters
        ----------
        gateway : ``DiscordGateway`` or ``DiscordGatewayVoice``
            The gateway of kokoro.
        
        Returns
        -------
        kokoro : ``Kokoro``
        
        Notes
        -----
        Before the coroutine returns a full loop is done to ensure `._start` is already started up.
        """
        self = object.__new__(cls)
        
        self.gateway = gateway
        self.interval = 40.0 # we set it from outside
        self.running = False
        now_ = perf_counter()
        self.last_answer = now_
        self.last_send = now_
        self.latency = self.DEFAULT_LATENCY
        self.ws_waiter = None
        self.beater = None
        self.beat_task = None
        self.beat_waiter = None
        self.task = Task(self._start(), KOKORO)
        
        #skip 1 loop
        future = Future(KOKORO)
        future.set_result(None)
        await future
        
        return self
    
    async def restart(self):
        """
        Restarts kokoro.
        
        This method is a coroutine.
        """
        self.cancel()
        future = Future(KOKORO)
        future.set_result(None)
        await future #skip 1 loop
        self.task = Task(self._start(), KOKORO)
        await future #skip 1 loop
    
    async def _start(self):
        """
        The main keep alive task of kokoro.
        
        The control flow of the method is the following:
            - Set `.ws_waiter` and wait till it is done. If it is done, means that kokoro's gateway is connected.
                If it was cancelled meanwhile, means the beat task should stop.
            - Sets `.beater` to ``._keep_beating`` and awaits it. This is the main beater of kokoro and runs meanwhile
                it's gateway is connected. This task can be cancelled, but we ignore that case.
            - If `.running` is still `True`, repeat.
        
        This method is a coroutine.
        """
        self.running = True
        while True:
            #wait for start
            try:
                waiter = Future(KOKORO)
                self.ws_waiter = waiter
                await waiter
            except CancelledError:
                #kokoro cancelled, client shuts down
                break
            finally:
                self.ws_waiter = None
            
            #keep beating
            try:
                beater = Task(self._keep_beating(), KOKORO)
                self.beater = beater
                await beater
            except CancelledError:
                #connection cancelled, lets wait for it
                pass
            finally:
                #make sure
                self.beater = None
            
            if self.running:
                continue
            
            break
        
        self.task = None
     
    async def _keep_beating(self):
        """
        The beater task of kokoro.
        
        The control flow of tze method is the following:
            - If `.should_beat` is not `True`, break out.
            - Wait `.interval` time and set the waiting ``Future`` to `self.beat_waiter`. If kokoro is cancelled, or
                if it should beat now, it is cancelled and we repeat the loop.
            - If we did not get answer since last beating (what is triggered first from outside), then we stop the
                gateway. We also break out from the loop to terminate the beating state and we will wait for the
                websocket to connect again.
            - We beat one with starting `gateway._beat` as a ``Task`` and setting it to `.beat_task`. If the task is
                not completed before it's respective timeout, we stop the gateway here as well. We also break out
                from the loop to terminate the beating state and we will wait for the websocket to connect again.
                This task can also be cancelled. If cancellation occurs, we repeat the loop.
            - If the beating task is done, we update `.last_send` to the current `perf_counter` time. Repeat the loop.
        
        This method is a coroutine.
        """
        self.last_answer = perf_counter()
        gateway = self.gateway
        self.should_beat = True
        while self.should_beat:
            waiter = sleep(self.interval, KOKORO)
            self.beat_waiter = waiter
            try:
                await waiter
            except CancelledError:
                self.last_send = perf_counter()
                continue
            
            self.beat_waiter = None
            
            if (self.last_answer+self.interval+HEARTBEAT_TIMEOUT)-perf_counter() <= 0.0:
                self.should_beat = False
                Task(gateway.terminate(), KOKORO)
                break
            
            try:
                task = Task(gateway._beat(), KOKORO)
                future_or_timeout(task, HEARTBEAT_TIMEOUT)
                self.beat_task = task
                await task
            except TimeoutError:
                self.should_beat = False
                Task(gateway.terminate(), KOKORO)
                break
            except CancelledError:
                continue
            finally:
                self.beat_task = None
            
            self.last_send = perf_counter()
    
    def start_beating(self):
        """
        Starts kokoro's beating. Handles all the cases around the board to do so.
        """
        # case 1 : we are not running
        if not self.running:
            Task(self._start_beating(), KOKORO)
            return
        
        #case 2 : we wait for ws
        waiter = self.ws_waiter
        if waiter is not None:
            waiter.set_result(None)
            return
        
        #case 3 : we wait for beat response
        waiter = self.beat_waiter
        if waiter is not None:
            waiter.cancel()
            return
        
        #case 4 : we are beating
        task = self.beat_task
        if task is not None:
            task.cancel()
            return
    
    async def _start_beating(self):
        """
        Internal method to start beating when kokoro is not running.
        
        This method is a coroutine.
        """
        #starts kokoro, then beating
        self.task = Task(self._start(), KOKORO)
        #skip 1 loop
        future = Future(KOKORO)
        future.set_result(None)
        await future
        
        waiter = self.ws_waiter
        if waiter is not None:
            waiter.set_result(None)
    
    def answered(self):
        """
        When gateway received answer after beating, it's gateway calls this method.
        
        This method updates `.last_answer` and `.latency` as well. Because of the updated `.last_answer`,
        ``._keep_beating`` task will know that it's gateway is still connected.
        """
        now_ = perf_counter()
        self.last_answer = now_
        self.latency = now_-self.last_send
    
    def terminate(self):
        """
        Terminates kokoro till restart. Called when it's gateway will resume it's connection.
        """
        #case 1 : we are not running
        if not self.running:
            return
        
        #case 2 : we are waiting for ws
        waiter = self.ws_waiter
        if waiter is not None:
            #it is fine, that is what we should do
            return
        
        #case 2 : we are beating
        beater = self.beater
        if beater is not None:
            self.should_beat = False
            
            #case 3.1 : we wait to beat
            waiter = self.beat_waiter
            if waiter is not None:
                waiter.cancel()
                return
            
            #case 3.2: we are beating
            task = self.beat_task
            if task is not None:
                task.cancel()
                return
            
    def cancel(self):
        """
        Stops kokoro. Called when it's gateway is closing.
        """
        #case 1 : we are not running
        if not self.running:
            return
        self.running = False
        
        #case 2 : we are waiting for ws
        waiter = self.ws_waiter
        if (waiter is not None):
            waiter.cancel()
            return
        
        #case 3 : we are beating
        beater = self.beater
        if beater is not None:
            self.should_beat = False
            
            #case 3.1 : we wait to beat
            waiter = self.beat_waiter
            if waiter is not None:
                waiter.cancel()
                return
            
            #case 3.2: we are beating
            task = self.beat_task
            if task is not None:
                task.cancel()
                return
            
    async def beat_now(self):
        """
        Sends a heartbeat event now. If kokoro is not at beating state, or anywhere in between, resets itself as it just
        received a heartbeat answer right now.
        
        This method is a coroutine.
        """
        while True: #GOTO
            #case 1 : we are not running:
            if not self.running:
                Task(self._start_beating(), KOKORO)
                should_beat_now = True # True = send beat data
                break
            
            # case 2 : we wait for ws
            waiter = self.ws_waiter
            if waiter is not None:
                waiter.set_result(None)
                should_beat_now = True # True = send beat data
                break
            
            # case 3 : we wait to beat
            waiter = self.beat_waiter
            if waiter is not None:
                # better skip a beat
                waiter.cancel()
                should_beat_now = True # True = send beat data
                break
            
            # case 4 : we already beat
##            task=self.beat_task
##            if task is not None:
##                should_beat_now=False
##                break
            
            should_beat_now = False
            break
        
        if should_beat_now:
            await self.gateway._beat()
        else:
            self.answered()
            
    if __debug__:
        def __del__(self):
            # bug?
            waiter = self.ws_waiter
            if waiter is not None:
                waiter.__silence__()
            
            # despair?
            waiter = self.beat_waiter
            if waiter is not None:
                waiter.__silence__()
            
    def __repr__(self):
        """Returns kokoro's representation."""
        result = [
            '<',
            self.__class__.__name__,
            ' client=',
            repr(self.gateway.client.full_name),
            ' state=',
                ]
        
        if self.task is None:
            state = 'stopped'
        elif self.beater is None:
            state = 'waiting for websocket'
        else:
            state = 'beating'
        
        result.append(state)
        result.append('>')
        
        return ''.join(result)


del WeakValueDictionary
del ClientDictionary
del sys



# -*- coding: utf-8 -*-
__all__ = ('ChooseMenu', 'Closer', 'Cooldown', 'GUI_STATE_CANCELLED', 'GUI_STATE_CANCELLING', 'GUI_STATE_READY',
    'GUI_STATE_SWITCHING_CTX', 'GUI_STATE_SWITCHING_PAGE', 'Timeouter', 'Pagination', 'WaitAndContinue',
    'ReactionAddWaitfor', 'ReactionDeleteWaitfor', 'multievent', 'wait_for_message', 'wait_for_reaction', )

from ...backend.futures import Task, Future
from ...backend.event_loop import LOOP_TIME

from ...discord.parsers import EventWaitforBase
from ...discord.emoji import BUILTIN_EMOJIS
from ...discord.exceptions import DiscordException, ERROR_CODES
from ...discord.client_core import KOKORO
from ...discord.embed import Embed
from ...discord.channel import ChannelTextBase
from ...discord.message import Message

from .command import CommandWrapper

class ReactionAddWaitfor(EventWaitforBase):
    """
    Implements waiting for `reaction_add` events.
    
    Attributes
    ----------
    waitfors : `WeakValueDictionary` of (``DiscordEntity``, `async-callable`) items
        An auto-added container to store `entity` - `async-callable` pairs.
    
    Class Attributes
    ----------------
    __event_name__ : `str` = `'reaction_add'`
        Predefined name to what the event handler will be added.
    """
    __slots__ = ()
    __event_name__ = 'reaction_add'

class ReactionDeleteWaitfor(EventWaitforBase):
    """
    Implements waiting for `reaction_delete` events.
    
    Attributes
    ----------
    waitfors : `WeakValueDictionary` of (``DiscordEntity``, `async-callable`) items
        An auto-added container to store `entity` - `async-callable` pairs.
    
    Class Attributes
    ----------------
    __event_name__ : `str` = `'reaction_delete'`
        Predefined name to what the event handler will be added.
    """
    __slots__ = ()
    __event_name__ = 'reaction_delete'

class multievent(object):
    """
    Helper class to hold more waitfor event handlers together allowing to add `target` - `waiter` pairs at the same to
    more.
    
    Attributes
    ----------
    events : `tuple` of `Any`
        A `tuple` of the contained event handlers.
    """
    __slots__=('events',)
    
    def __init__(self, *events):
        """
        Creates a `multievent` instance with the given event handlers
        
        Parameters
        ----------
        *events : `Any`
            The event handlers to hold together.
        """
        self.events = events
    
    def append(self, target, waiter):
        """
        Adds the given `target` - `waiter` pair to the contained event handlers.
        
        Parameters
        ----------
        target : ``DiscordEntity`` instance
        waiter : `async callable`
        """
        for event in self.events:
            event.append(target, waiter)
    
    def remove(self, target, waiter):
        """
        Removes the given `target` - `waiter` pair to the contained event handlers.
        
        Parameters
        ----------
        target : ``DiscordEntity`` instance
        waiter : `async callable`
        """
        for event in self.events:
            event.remove(target, waiter)

class Timeouter(object):
    """
    Executes timing out feature on ``Pagination`` and on other familiar types.
    
    Attributes
    ----------
    handle : `None` or ``TimerHandle``
        Handle to wake_up the timeouter with it's `.__step` function.
        Set to `None`, when the respective timeout is over or if the timeout is cancelled.
    owner : `Any`
        The object what uses the timeouter.
        Set to `None`, when the respective timeout is over or if the timeout is cancelled.
    timeout : `float`
        The time with what the timeout will be expired when it's current waiting cycle is over.
    """
    __slots__ = ('handle', 'owner', 'timeout')
    def __init__(self, owner, timeout):
        """
        Creates a new ``Timeouter`` instance with the given `owner` and `timeout`.
        
        Parameters
        ----------
        owner : `Any`
            The object what uses the timeouter.
        timeout : `float`
            The time with what the timeout will be expired when it's current waiting cycle is over.
        """
        self.owner = owner
        self.timeout = 0.0
        self.handle = KOKORO.call_later(timeout, self.__step, self)
    
    @staticmethod
    def __step(self):
        """
        Executes a timeouter cycle.
        
        Increases the timeout if ``.timeout`` was updated. If not and applicable, calls it's ``.owner``'s
        `.canceller` with `TimeoutError` and unlinks ``.owner`` and `owner.canceller`,
        """
        timeout = self.timeout
        if timeout > 0.0:
            self.handle = KOKORO.call_later(timeout, self.__step, self)
            self.timeout = 0.0
            return
        
        self.handle = None
        owner = self.owner
        if owner is None:
            return
        
        self.owner = None
        
        canceller = owner.canceller
        if canceller is None:
            return
        
        owner.canceller = None
        Task(canceller(owner, TimeoutError()), KOKORO)
    
    def cancel(self):
        """
        Cancels the timeouter.
        
        Should be called by the timeouter's owner when it is cancelled with an other exception.
        """
        handle = self.handle
        if handle is None:
            return
        
        self.handle = None
        handle.cancel()
        self.owner = None
    
    def set_timeout(self, value):
        """
        Sets the timeouter of the timeouter to the given value.
        """
        handle = self.handle
        if handle is None:
            # Cannot change timeout of expired timeouter
            return
        
        if value <= 0.0:
            self.timeout = 0.0
            handle._run()
            handle.cancel()
            return
        
        now = LOOP_TIME()
        next_step = self.handle.when
        
        planed_end = now+value
        if planed_end < next_step:
            handle.cancel()
            self.handle = KOKORO.call_at(planed_end, self.__step, self)
            return
        
        self.timeout = planed_end-next_step
    
    def get_expiration_delay(self):
        """
        Returns after how much time the timeouter will expire.
        
        If the timeouter already expired, returns `0.0˙.
        
        Returns
        -------
        time_left : `float`
        """
        handle = self.handle
        if handle is None:
            return 0.0
        
        return handle.when-LOOP_TIME()+self.timeout

GUI_STATE_READY          = 0
GUI_STATE_SWITCHING_PAGE = 1
GUI_STATE_CANCELLING     = 2
GUI_STATE_CANCELLED      = 3
GUI_STATE_SWITCHING_CTX  = 4

class Pagination(object):
    """
    A builtin option to display paginated messages, allowing the users moving between the pages with arrow emojis.
    
    The class allows modifications and closing it's representations for every user. Also works at private channels.
    
    Picks up on reaction additions and on reaction deletions as well and removes the added reactions on if has
    permission, which might be missing, like in DM-s.
    
    Attributes
    ----------
    canceller : `None` or `function`
        The function called when the ``Pagination`` is cancelled or when it expires. This is a onetime use and after
        it was used, is set as `None`.
    channel : ``ChannelTextBase`` instance
        The channel where the ``Pagination`` is executed.
    check : `None` or `callable`
        A callable what decides whether the ``Pagination`` should process a received reaction event. Defaults to
        `None`.
        
        Should accept the following parameters:
        +-----------+---------------------------------------------------+
        | Name      | Type                                              |
        +===========+===================================================+
        | event     | ``ReactionAddEvent`` or ``ReactionDeleteEvent``   |
        +-----------+---------------------------------------------------+
        
        Note, that ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
        
        Should return the following values:
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | should_process    | `bool`    |
        +-------------------+-----------+
    
    client : ``Client`` of ``Embed`` (or any compatible)
        The client who executes the ``Pagination``.
    message : `None` or ``Message``
        The message on what the ``Pagination`` is executed.
    page : `int`
        The current page's index.
    pages : `indexable`
        An indexable container, what stores the displayable ``Embed``-s.
    task_flag : `int`
        A flag to store the state of the ``Pagination``.
        
        Possible values:
        +---------------------------+-------+-----------------------------------------------------------------------+
        | Respective name           | Value | Description                                                           |
        +===========================+=======+=======================================================================+
        | GUI_STATE_READY           | 0     | The Pagination does nothing, is ready to be used.                     |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_SWITCHING_PAGE  | 1     | The Pagination is currently changing it's page.                       |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_CANCELLING      | 2     | The pagination is currently changing it's page, but it was cancelled  |
        |                           |       | meanwhile.                                                            |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_CANCELLED       | 3     | The pagination is, or is being cancelled right now.                   |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_SWITCHING_CTX   | 4     | The Pagination is switching context. Not used by the default class,   |
        |                           |       | but expected.                                                         |
        +---------------------------+-------+-----------------------------------------------------------------------+
    timeout : `float`
        The timeout of the ``Pagination`` in seconds.
    timeouter : `None` or ``Timeouter``
        Executes the timing out feature on the ``Pagination``.
    
    Class Attributes
    ----------------
    LEFT2 : ``Emoji`` = `BUILTIN_EMOJIS['track_previous']`
        The emoji used to move to the first page.
    LEFT : ``Emoji`` = `BUILTIN_EMOJIS['arrow_backward']`
        The emoji used to move to the previous page.
    RIGHT : ``Emoji`` = `BUILTIN_EMOJIS['arrow_forward']`
        The emoji used to move on the next page.
    RIGHT2 : ``Emoji`` = `BUILTIN_EMOJIS['track_next']`
        The emoji used to move on the last page.
    CANCEL : ``Emoji`` = `BUILTIN_EMOJIS['x']`
        The emoji used to cancel the ``Pagination``.
    EMOJIS : `tuple` (`Emoji`, `Emoji`, `Emoji`, `Emoji`, `Emoji`) = `(LEFT2, LEFT, RIGHT, RIGHT2, CANCEL,)`
        The emojis to add on the respective message in order.
    """
    LEFT2  = BUILTIN_EMOJIS['track_previous']
    LEFT   = BUILTIN_EMOJIS['arrow_backward']
    RIGHT  = BUILTIN_EMOJIS['arrow_forward']
    RIGHT2 = BUILTIN_EMOJIS['track_next']
    CANCEL = BUILTIN_EMOJIS['x']
    EMOJIS = (LEFT2, LEFT, RIGHT, RIGHT2, CANCEL,)
    
    __slots__ = ('canceller', 'channel', 'check', 'client', 'message', 'page', 'pages', 'task_flag', 'timeout', 'timeouter')
    
    async def __new__(cls, client, channel, pages, timeout=240., message=None, check=None):
        """
        Creates a new pagination with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who will execute the ``Pagination``.
        channel : ``ChannelTextBase`` instance or ``Message``
            The channel where the ``Pagination`` will be executed. Pass it as a ``Message`` instance to send a reply.
        pages : `indexable-container`
            An indexable container, what stores the displayable pages.
        timeout : `float`, Optional
            The timeout of the ``Pagination`` in seconds. Defaults to `240.0`.
        message : `None` or ``Message``, Optional
            The message on what the ``Pagination`` will be executed. If not given a new message will be created.
            Defaults to `None`.
        check : `None` or `callable`, Optional
            A callable what decides whether the ``Pagination`` should process a received reaction event. Defaults to
            `None`.
            
            Should accept the following parameters:
            +-----------+---------------------------------------------------+
            | Name      | Type                                              |
            +===========+===================================================+
            | event     | ``ReactionAddEvent`` or ``ReactionDeleteEvent``   |
            +-----------+---------------------------------------------------+
            
            Note, that ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
            
            Should return the following values:
            +-------------------+-----------+
            | Name              | Type      |
            +===================+===========+
            | should_process    | `bool`    |
            +-------------------+-----------+
        
        Returns
        -------
        self : `None` or ``Pagination``
            If `pages` is an empty container, returns `None`.
        
        Raises
        ------
        TypeError
            `channel`'s type is incorrect.
        """
        if not pages:
            return None
        
        if isinstance(channel, ChannelTextBase):
            target_channel = channel
        elif isinstance(channel, Message):
            target_channel = channel.channel
        else:
            raise TypeError('`channel` can be given only as `ChannelTextBase` or as `Message` instance.')
        
        self = object.__new__(cls)
        self.check = check
        self.client = client
        self.channel = target_channel
        self.pages = pages
        self.page = 0
        self.canceller = cls._canceller
        self.task_flag = GUI_STATE_READY
        self.message = message
        self.timeout = timeout
        self.timeouter = None
        
        try:
            if message is None:
                message = await client.message_create(channel, pages[0])
                self.message = message
            else:
                await client.message_edit(message, pages[0])
            
            if not target_channel.cached_permissions_for(client).can_add_reactions:
                return self
            
            if len(self.pages)>1:
                for emoji in self.EMOJIS:
                    await client.reaction_add(message, emoji)
            else:
                await client.reaction_add(message, self.CANCEL)
        except BaseException as err:
            if isinstance(err, ConnectionError):
                return None
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_message, # message deleted
                        ERROR_CODES.unknown_channel, # message's channel deleted
                        ERROR_CODES.max_reactions, # reached reaction 20, some1 is trolling us.
                        ERROR_CODES.invalid_access, # client removed
                        ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                        ERROR_CODES.cannot_message_user, # user has dm-s disallowed
                            ):
                    return None
            
            raise
        
        self.timeouter = Timeouter(self, timeout=timeout)
        client.events.reaction_add.append(message, self)
        client.events.reaction_delete.append(message, self)
        return self
    
    async def __call__(self, client, event):
        """
        Called when a reaction is added or removed from the respective message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who executes the ``Pagination``
        event : ``ReactionAddEvent``, ``ReactionDeleteEvent``
            The received event.
        """
        if event.user.is_bot:
            return
        
        if (event.emoji not in self.EMOJIS):
            return
        
        if (event.delete_reaction_with(client) == event.DELETE_REACTION_NOT_ADDED):
            return
        
        check = self.check
        if (check is not None):
            try:
                should_continue = check(event)
            except BaseException as err:
                await client.events.error(client, f'{self!r}.__call__', err)
                return
            
            if not should_continue:
                return
        
        emoji = event.emoji
        task_flag = self.task_flag
        if task_flag != GUI_STATE_READY:
            if task_flag == GUI_STATE_SWITCHING_PAGE:
                if emoji is self.CANCEL:
                    self.task_flag = GUI_STATE_CANCELLING
                return
            
            # ignore GUI_STATE_CANCELLED and GUI_STATE_SWITCHING_CTX
            return
        
        while True:
            if emoji is self.LEFT:
                page = self.page-1
                break
            
            if emoji is self.RIGHT:
                page = self.page+1
                break
            
            if emoji is self.CANCEL:
                self.task_flag = GUI_STATE_CANCELLED
                try:
                    await client.message_delete(self.message)
                except BaseException as err:
                    self.cancel()
                    
                    if isinstance(err, ConnectionError):
                        # no internet
                        return
                    
                    if isinstance(err, DiscordException):
                        if err.code in (
                                ERROR_CODES.unknown_channel, # message's channel deleted
                                ERROR_CODES.invalid_access, # client removed
                                    ):
                            return
                    
                    await client.events.error(client, f'{self!r}.__call__', err)
                    return
                
                else:
                    self.cancel()
                    return
            
            if emoji is self.LEFT2:
                page = 0
                break
            
            if emoji is self.RIGHT2:
                page = len(self.pages)-1
                break
            
            return
        
        if page < 0:
            page = 0
        elif page >= len(self.pages):
            page = len(self.pages)-1
        
        if self.page == page:
            return
        
        self.page = page
        self.task_flag = GUI_STATE_SWITCHING_PAGE
        
        try:
            await client.message_edit(self.message, self.pages[page])
        except BaseException as err:
            self.task_flag = GUI_STATE_CANCELLED
            self.cancel()
            
            if isinstance(err, ConnectionError):
                # no internet
                return
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_message, # message deleted
                        ERROR_CODES.unknown_channel, # channel deleted
                        ERROR_CODES.invalid_access, # client removed
                            ):
                    return
            
            # We definitely do not want to silence `ERROR_CODES.invalid_form_body`
            await client.events.error(client, f'{self!r}.__call__', err)
            return
        
        if self.task_flag == GUI_STATE_CANCELLING:
            self.task_flag = GUI_STATE_CANCELLED
            self.cancel()
            
            try:
                await client.message_delete(self.message)
            except BaseException as err:
                
                if isinstance(err, ConnectionError):
                    # no internet
                    return
                
                if isinstance(err, DiscordException):
                    if err.code in (
                            ERROR_CODES.unknown_channel, #message's channel deleted
                            ERROR_CODES.invalid_access, # client removed
                                ):
                        return
                
                await client.events.error(client, f'{self!r}.__call__', err)
                return
            
            return
            
        self.task_flag = GUI_STATE_READY
        self.timeouter.set_timeout(self.timeout)
    
    async def _canceller(self, exception,):
        """
        Used when the ``Pagination`` is cancelled.
        
        First of all removes the pagination from waitfors, so it will not wait for reaction events, then sets the
        ``.task_flag`` of the it to `GUI_STATE_CANCELLED`.
        
        If `exception` is given as `TimeoutError`, then removes the ``Pagination``'s reactions from the respective
        message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        exception : `None` or ``BaseException`` instance
            Exception to cancel the ``Pagination`` with.
        """
        client = self.client
        message = self.message
        
        client.events.reaction_add.remove(message, self)
        client.events.reaction_delete.remove(message, self)
        
        if self.task_flag == GUI_STATE_SWITCHING_CTX:
            # the message is not our, we should not do anything with it.
            return
        
        self.task_flag = GUI_STATE_CANCELLED
        
        if exception is None:
            return
        
        if isinstance(exception, TimeoutError):
            if self.channel.cached_permissions_for(client).can_manage_messages:
                try:
                    await client.reaction_clear(message)
                except BaseException as err:
                    
                    if isinstance(err,ConnectionError):
                        # no internet
                        return
                    
                    if isinstance(err,DiscordException):
                        if err.code in (
                                ERROR_CODES.unknown_message, # message deleted
                                ERROR_CODES.unknown_channel, # channel deleted
                                ERROR_CODES.invalid_access, # client removed
                                ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                                    ):
                            return
                    
                    await client.events.error(client, f'{self!r}._canceller', err)
                    return
            return
        
        timeouter = self.timeouter
        if (timeouter is not None):
            timeouter.cancel()
        #we do nothing
    
    def cancel(self, exception=None):
        """
        Cancels the pagination, if it is not cancelled yet.
        
        Parameters
        ----------
        exception : `None` or ``BaseException`` instance, Optional
            Exception to cancel the pagination with. Defaults to `None`
        """
        canceller = self.canceller
        if canceller is None:
            return
        
        self.canceller = None
        
        timeouter = self.timeouter
        if (timeouter is not None):
            timeouter.cancel()
        
        return Task(canceller(self, exception), KOKORO)
    
    def __repr__(self):
        """Returns the pagination's representation."""
        result = [
            '<', self.__class__.__name__,
            ' client=', repr(self.client),
            ', pages=', repr(len(self.pages)),
            ', page=', repr(self.page),
            ', channel=', repr(self.channel),
            ', task_flag='
                ]
        
        task_flag = self.task_flag
        result.append(repr(task_flag))
        result.append(' (')
        
        task_flag_name = (
            'GUI_STATE_READY',
            'GUI_STATE_SWITCHING_PAGE',
            'GUI_STATE_CANCELLING',
            'GUI_STATE_CANCELLED',
            'GUI_STATE_SWITCHING_CTX',
                )[task_flag]
        
        result.append(task_flag_name)
        result.append(')>')
        
        return ''.join(result)

class Closer(object):
    """
    Familiar to ``Pagination``, but can be used if the given content has only 1 page, so only an `x` would show up.
    
    Picks up on reaction additions by any users.
    
    Attributes
    ----------
    canceller : `None` or `function`
        The function called when the ``Closer`` is cancelled or when it expires. This is a onetime use and after
        it was used, is set as `None`.
    channel : ``ChannelTextBase`` instance
        The channel where the ``Closer`` is executed.
    check : `None` or `callable`
        A callable what decides whether the ``Closer`` should process a received reaction event. Defaults to
        `None`.
        
        Should accept the following parameters:
        +-----------+---------------------------------------------------+
        | Name      | Type                                              |
        +===========+===================================================+
        | event     | ``ReactionAddEvent`` or ``ReactionDeleteEvent``   |
        +-----------+---------------------------------------------------+
        
        Note, that ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
        
        Should return the following values:
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | should_process    | `bool`    |
        +-------------------+-----------+
    client : ``Client`` of ``Embed`` (or any compatible)
        The client who executes the ``Closer``.
    message : `None` or ``Message``
        The message on what the ``Closer`` is executed.
    task_flag : `int`
        A flag to store the state of the ``Closer``.
        
        Possible values:
        +---------------------------+-------+-----------------------------------------------------------------------+
        | Respective name           | Value | Description                                                           |
        +===========================+=======+=======================================================================+
        | GUI_STATE_READY           | 0     | The closer does nothing, is ready to be used.                         |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_SWITCHING_PAGE  | 1     | The closer is currently changing it's page.                           |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_CANCELLING      | 2     | The pagination is currently changing it's page, but it was cancelled  |
        |                           |       | meanwhile.                                                            |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_CANCELLED       | 3     | The pagination is, or is being cancelled right now.                   |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_SWITCHING_CTX   | 4     | The closer is switching context. Not used by the default class,       |
        |                           |       | but expected.                                                         |
        +---------------------------+-------+-----------------------------------------------------------------------+
    timeouter : `None` or ``Timeouter``
        Executes the timing out feature on the ``Closer``.
    
    Class Attributes
    ----------------
    CANCEL : ``Emoji`` = `BUILTIN_EMOJIS['x']`
        The emoji used to cancel the ``Closer``.
    """
    CANCEL = BUILTIN_EMOJIS['x']
    
    __slots__ = ('canceller', 'channel', 'check', 'client', 'message', 'task_flag', 'timeouter')
    
    async def __new__(cls, client, channel, content, timeout=240., message=None, check=None):
        """
        Creates a new pagination with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who will execute the ``Closer``.
        channel : ``ChannelTextBase`` instance or ``Message``
            The channel where the ``Closer`` will be executed.  Pass it as a ``Message`` instance to send a reply.
        content : ``Any`
            The displayed content.
        timeout : `float`, Optional
            The timeout of the ``Closer`` in seconds. Defaults to `240.0`.
        message : `None` or ``Message``, Optional
            The message on what the ``Closer`` will be executed. If not given a new message will be created.
            Defaults to `None`.
        check : `None` or `callable`, Optional
            A callable what decides whether the ``Closer`` should process a received reaction event. Defaults to
            `None`.
            
            Should accept the following parameters:
            +-----------+---------------------------------------------------+
            | Name      | Type                                              |
            +===========+===================================================+
            | event     | ``ReactionAddEvent`` or ``ReactionDeleteEvent``   |
            +-----------+---------------------------------------------------+
            
            Note, that ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
            
            Should return the following values:
            +-------------------+-----------+
            | Name              | Type      |
            +===================+===========+
            | should_process    | `bool`    |
            +-------------------+-----------+
        
        Returns
        -------
        self : `None` or ``Closer``
        
        Raises
        ------
        TypeError
            `channel`'s type is incorrect.
        """
        if isinstance(channel, ChannelTextBase):
            target_channel = channel
        elif isinstance(channel, Message):
            target_channel = channel.channel
        else:
            raise TypeError('`channel` can be given only as `ChannelTextBase` or as `Message` instance.')
        
        self = object.__new__(cls)
        self.check = check
        self.client = client
        self.channel = target_channel
        self.canceller = cls._canceller
        self.task_flag = GUI_STATE_READY
        self.message = message
        self.timeouter = None
        
        try:
            if message is None:
                message = await client.message_create(channel, content)
                self.message = message
            else:
                await client.message_edit(message, content)
            
            if not target_channel.cached_permissions_for(client).can_add_reactions:
                return self
            
            await client.reaction_add(message, self.CANCEL)
        except BaseException as err:
            if isinstance(err, ConnectionError):
                return None
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_message, # message deleted
                        ERROR_CODES.unknown_channel, # message's channel deleted
                        ERROR_CODES.max_reactions, # reached reaction 20, some1 is trolling us.
                        ERROR_CODES.invalid_access, # client removed
                        ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                        ERROR_CODES.cannot_message_user, # user has dm-s disallowed
                            ):
                    return None
            
            raise
        
        self.timeouter = Timeouter(self, timeout=timeout)
        client.events.reaction_add.append(message, self)
        return self
    
    
    async def __call__(self, client, event):
        """
        Called when a reaction is added on the respective message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who executes the ``Closer``
        event : ``ReactionAddEvent``
            The received event.
        """
        if event.user.is_bot:
            return
        
        if (event.emoji is not self.CANCEL):
            return
        
        task_flag = self.task_flag
        if task_flag != GUI_STATE_READY:
            # ignore GUI_STATE_SWITCHING_PAGE and GUI_STATE_CANCELLED and GUI_STATE_SWITCHING_CTX
            return
        
        check = self.check
        if (check is not None):
            try:
                should_continue = check(event)
            except BaseException as err:
                await client.events.error(client, f'{self!r}.__call__', err)
                return
            
            if not should_continue:
                return
        
        self.task_flag = GUI_STATE_CANCELLED
        try:
            await client.message_delete(self.message)
        except BaseException as err:
            self.cancel()
            
            if isinstance(err, ConnectionError):
                # no internet
                return
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_channel, # message's channel deleted
                        ERROR_CODES.unknown_message, # message deleted
                        ERROR_CODES.invalid_access, # client removed
                            ):
                    return
            
            await client.events.error(client, f'{self!r}.__call__', err)
            return
    
    async def _canceller(self, exception,):
        """
        Used when the ``Closer`` is cancelled.
        
        First removes the respective waitfors of the canceller, then set it's ``.task_flag`` of the it to
        `GUI_STATE_CANCELLED`.
        
        If `exception` is given as `TimeoutError`, then removes the ``Closer``'s reactions from the respective
        message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        exception : `None` or ``BaseException`` instance
            Exception to cancel the ``Closer`` with.
        """
        client = self.client
        message = self.message
        
        client.events.reaction_add.remove(message, self)
        
        if self.task_flag == GUI_STATE_SWITCHING_CTX:
            # the message is not our, we should not do anything with it.
            return
        
        self.task_flag = GUI_STATE_CANCELLED
        
        if exception is None:
            return
        
        if isinstance(exception, TimeoutError):
            if self.channel.cached_permissions_for(client).can_manage_messages:
                try:
                    await client.reaction_clear(message)
                except BaseException as err:
                    
                    if isinstance(err, ConnectionError):
                        # no internet
                        return
                    
                    if isinstance(err, DiscordException):
                        if err.code in (
                                ERROR_CODES.unknown_message, # message deleted
                                ERROR_CODES.unknown_channel, # channel deleted
                                ERROR_CODES.invalid_access, # client removed
                                ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                                    ):
                            return
                    
                    await client.events.error(client, f'{self!r}._canceller', err)
                    return
            return
        
        timeouter = self.timeouter
        if (timeouter is not None):
            timeouter.cancel()
        #we do nothing
    
    def cancel(self, exception=None):
        """
        Cancels the closer, if it is not cancelled yet.
        
        Parameters
        ----------
        exception : `None` or ``BaseException`` instance, Optional
            Exception to cancel the closer with. Defaults to `None`
        """
        canceller = self.canceller
        if canceller is None:
            return
        
        self.canceller = None
        
        timeouter = self.timeouter
        if (timeouter is not None):
            timeouter.cancel()
        
        return Task(canceller(self, exception), KOKORO)
    
    def __repr__(self):
        """Returns the pagination's representation."""
        result = [
            '<', self.__class__.__name__,
            ' client=', repr(self.client),
            ', channel=', repr(self.channel),
            ', task_flag='
                ]
        
        task_flag = self.task_flag
        result.append(repr(task_flag))
        result.append(' (')
        
        task_flag_name = (
            'GUI_STATE_READY',
            'GUI_STATE_SWITCHING_PAGE',
            'GUI_STATE_CANCELLING',
            'GUI_STATE_CANCELLED',
            'GUI_STATE_SWITCHING_CTX',
                )[task_flag]
        
        result.append(task_flag_name)
        result.append(')>')
        
        return ''.join(result)


class ChooseMenu(object):
    """
    Familiar to ``Pagination``, but instead of just displaying multiple pages of text, it allows the user to select
    a displayed option.
    
    The class allows modifications and closing it's representations for every user. Also works at private channels.
    
    Picks up on reaction additions and on reaction deletions as well and removes the added reactions on if has
    permission, which might be missing, like in DM-s.
    
    Attributes
    ----------
    canceller : `None` or `function`
        The function called when the ``ChooseMenu`` is cancelled or when it expires. This is a onetime use and after
        it was used, is set as `None`.
    channel : ``ChannelTextBase`` instance
        The channel where the ``ChooseMenu`` is executed.
    check : `None` or `callable`
        A callable what decides whether the ``ChooseMenu`` should process a received reaction event. Defaults to
        `None`.
        
        Should accept the following parameters:
        +-----------+---------------------------------------------------+
        | Name      | Type                                              |
        +===========+===================================================+
        | event     | ``ReactionAddEvent`` or ``ReactionDeleteEvent``   |
        +-----------+---------------------------------------------------+
        
        Note, that ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
        
        Should return the following values:
        +-------------------+-----------+
        | Name              | Type      |
        +===================+===========+
        | should_process    | `bool`    |
        +-------------------+-----------+
    
    client : ``Client``
        The client who executes the ``ChooseMenu``.
    embed : ``Embed`` (or any compatible)
            An embed base, what's description and footer will be rendered with the given choices and with information
            about the respective page.
    message : `None` or ``Message``
        The message on what the ``ChooseMenu`` is executed.
    selected : `int`
        The currently selected option of the ``ChooseMenu``.
    choices : `indexable` of `Any`
        An indexable container, what stores the displayable choices.
        
        It's elements's type can be different from each other, and different structures act differently as well.
        There are the following cases:
        
        - If an element is `str` instance, then it will be used as an option's title and when selecting it, only that
            variable will be passed to the respective function when selected.
        
        - If an element is neither `str` or `tuple`, then it's `repr` will be used as an option's title, and only that
            variable will be passed to the respective function when selected.
        
        - If an element is `tuple` instance, then it's first element will be displayed as title. If it is `str`, then
            will be just simply added, however if not, then it's `repr` will be used. If selecting a `tuple` option,
            then it's element will be passed to the respective function.
    
    task_flag : `int`
        A flag to store the state of the ``ChooseMenu``.
        
        Possible values:
        +---------------------------+-------+-----------------------------------------------------------------------+
        | Respective name           | Value | Description                                                           |
        +===========================+=======+=======================================================================+
        | GUI_STATE_READY           | 0     | The ChooseMenu does nothing, is ready to be used.                     |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_SWITCHING_PAGE  | 1     | The ChooseMenu is currently changing it's page.                       |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_CANCELLING      | 2     | The ChooseMenu is currently changing it's page, but it was cancelled  |
        |                           |       | meanwhile.                                                            |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_CANCELLED       | 3     | The ChooseMenu is, or is being cancelled right now.                   |
        +---------------------------+-------+-----------------------------------------------------------------------+
        | GUI_STATE_SWITCHING_CTX   | 4     | The ChooseMenu is switching context. Not used by the default class,   |
        |                           |       | but expected.                                                         |
        +---------------------------+-------+-----------------------------------------------------------------------+
    timeout : `float`
        The timeout of the ``ChooseMenu`` in seconds.
    timeouter : `None` or ``Timeouter``
        Executes the timing out feature on the ``ChooseMenu``.
    prefix : `None` or `str`
        A prefix displayed before each option.
    selector : `async-callable`
        An `async-callable`, what is ensured when an option is selected.
        
        If the ``ChooseMenu`` is created only with `1` option, then it is ensured initially instead of creating the
        ``ChooseMenu`` itself. At this case, if `message` was not given (or given as `None`), then the `message`
        passed to the `selector` will be `None` as well.
        
        At least 3 parameters are passed to the `selector`:
        +-------------------+-------------------------------+
        | Respective name   | Type                          |
        +===================+===============================+
        | client            | ``Client``                    |
        +-------------------+-------------------------------+
        | channel           | ``ChannelTextBase`` instance  |
        +-------------------+-------------------------------+
        | message           | ``Message`` or `None`         |
        +-------------------+-------------------------------+
        
        The rest of the parameters depend on the respective choice (an elements of ``choices``). If the element is a
        `tuple` instance, then it's element will be passed, however if the choice is any other type, then only that
        object will be passed.
    
    Class Attributes
    ----------------
    UP : ``Emoji`` = `BUILTIN_EMOJIS['arrow_up_small']`
        The emoji used to move on the displayed option one above.
    DOWN : ``Emoji`` = `BUILTIN_EMOJIS['arrow_down_small']`
        The emoji used to move on the displayed option one under.
    LEFT : ``Emoji`` = `BUILTIN_EMOJIS['arrow_backward']`
        The emoji used to move on the previous page.
    RIGHT : ``Emoji`` = `BUILTIN_EMOJIS['arrow_forward']`
        The emoji used to move on the next page.
    SELECT : ``Emoji`` = `BUILTIN_EMOJIS['ok']`
        The emoji used to select an option.
    CANCEL : ``Emoji`` = `BUILTIN_EMOJIS['x']`
        The emoji used to cancel the ``ChooseMenu``.
    EMOJIS_RESTRICTED : `tuple` (`Emoji`, `Emoji`, `Emoji`, `Emoji`) = `(UP, DOWN, SELECT, CANCEL)`
        Restricted emojis, added when the choose menu has only options for 1 page.
    EMOJIS : `tuple` (`Emoji`, `Emoji`, `Emoji`, `Emoji`, `Emoji`, `Emoji`) = `(UP, DOWN, LEFT, RIGHT, SELECT, CANCEL)`
        Emojis added to the choose menu.
    """
    UP     = BUILTIN_EMOJIS['arrow_up_small']
    DOWN   = BUILTIN_EMOJIS['arrow_down_small']
    LEFT   = BUILTIN_EMOJIS['arrow_backward']
    RIGHT  = BUILTIN_EMOJIS['arrow_forward']
    SELECT = BUILTIN_EMOJIS['ok']
    CANCEL = BUILTIN_EMOJIS['x']
    EMOJIS_RESTRICTED = (UP, DOWN, SELECT, CANCEL)
    EMOJIS = (UP, DOWN, LEFT, RIGHT, SELECT, CANCEL)
    
    __slots__ = ('canceller', 'channel', 'check', 'client', 'embed', 'message', 'selected', 'choices', 'task_flag',
        'timeout', 'timeouter', 'prefix', 'selector')
    
    async def __new__(cls, client, channel, choices, selector, embed=Embed(), timeout=240., message=None, prefix=None,
            check=None):
        """
        Creates a new choose menu with the given parameters.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who executes the ``ChooseMenu``.
        channel : ``ChannelTextBase`` instance or ``Message``
            The channel where the ``ChooseMenu`` is executed. Pass it as a ``Message`` instance to send a reply.
        choices : `indexable` of `Any`
            An indexable container, what stores the displayable choices.
            
            It's elements's type can be different from each other, and different structures act differently as well.
            There are the following cases:
            
            - If an element is `str` instance, then it will be used as an option's title and when selecting it, only
                that variable will be passed to the respective function when selected.
            
            - If an element is neither `str` or `tuple`, then it's `repr` will be used as an option's title, and only
                that variable will be passed to the respective function when selected.
            
            - If an element is `tuple` instance, then it's first element will be displayed as title. If it is `str`,
                then will be just simply added, however if not, then it's `repr` will be used. If selecting a `tuple`
                option, then it's element will be passed to the respective function.
        selector : `async-callable`
            An `async-callable`, what is ensured when an option is selected.
            
            If the ``ChooseMenu`` is created only with `1` option, then it is ensured initially instead of creating
            the ``ChooseMenu`` itself. At this case, if `message` was not given (or given as `None`), then the
            `message` passed to the `selector` will be `None` as well.
            
            At least 3 parameters are passed to the `selector`:
            +-------------------+-------------------------------+
            | Respective name   | Type                          |
            +===================+===============================+
            | client            | ``Client``                    |
            +-------------------+-------------------------------+
            | channel           | ``ChannelTextBase`` instance  |
            +-------------------+-------------------------------+
            | message           | ``Message`` or `None`         |
            +-------------------+-------------------------------+
            
            The rest of the parameters depend on the respective choice (an elements of ``choices``). If the element is a
            `tuple` instance, then it's element will be passed, however if the choice is any other type, then only that
            object will be passed.
        embed : ``Embed`` (or any compatible)
            An embed base, what's description and footer will be rendered with the given choices and with information
            about the respective page. Defaults to an empty ``Embed`` instance.
        timeout : `float`, Optional
            The timeout of the ``ChooseMenu`` in seconds. Defaults to `240.0`.
        message : `None` or ``Message``, Optional
            The message on what the ``ChooseMenu`` will be executed. If not given a new message will be created.
            Defaults to `None`.
        prefix : `None` or `str`, Optional
            A prefix displayed before each option. Defaults to `None`.
        check : `None` or `callable`, Optional
            A callable what decides whether the ``ChooseMenu`` should process a received reaction event. Defaults to
            `None`.
            
            Should accept the following parameters:
            +-----------+---------------------------------------------------+
            | Name      | Type                                              |
            +===========+===================================================+
            | event     | ``ReactionAddEvent`` or ``ReactionDeleteEvent``   |
            +-----------+---------------------------------------------------+
            
            Note, that ``ReactionDeleteEvent`` is only given, when the client has no `manage_messages` permission.
            
            Should return the following values:
            +-------------------+-----------+
            | Name              | Type      |
            +===================+===========+
            | should_process    | `bool`    |
            +-------------------+-----------+
            
        Returns
        -------
        self : `None` or ``ChooseMenu``
            If `choices`'s length is less than `2`, then returns `None`.
        
        Raises
        ------
        TypeError
            `channel`'s type is incorrect.
        ValueError
            If `prefix` was not given as `None` and it's length is over `64` characters.
        """
        if (prefix is not None) and (len(prefix) > 100):
            raise ValueError(f'Please pass a prefix, what is shorter than 100 characters, got {prefix!r}.')
        
        if isinstance(channel, ChannelTextBase):
            target_channel = channel
        elif isinstance(channel, Message):
            target_channel = channel.channel
        else:
            raise TypeError('`channel` can be given only as `ChannelTextBase` or as `Message` instance.')
        
        result_ln = len(choices)
        if result_ln < 2:
            if result_ln == 1:
                choice = choices[0]
                if isinstance(choice, tuple):
                    coro = selector(client, target_channel, message, *choice)
                else:
                    coro = selector(client, target_channel, message, choice)
                await coro
            return None
        
        self = object.__new__(cls)
        self.check = check
        self.client = client
        self.channel = target_channel
        self.choices = choices
        self.selector = selector
        self.selected = 0
        self.canceller = cls._canceller
        self.task_flag = GUI_STATE_READY
        self.message = message
        self.timeout = timeout
        self.timeouter = None
        self.prefix = prefix
        self.embed = embed
        
        try:
            if message is None:
                message = await client.message_create(channel, embed=self._render_embed())
                self.message = message
            else:
                await client.message_edit(message, embed=self._render_embed())
            
            if not target_channel.cached_permissions_for(client).can_add_reactions:
                return self
            
            for emoji in (self.EMOJIS if (len(choices) > 10) else self.EMOJIS_RESTRICTED):
                await client.reaction_add(message, emoji)
        except BaseException as err:
            if isinstance(err, ConnectionError):
                return self
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_message, # message deleted
                        ERROR_CODES.unknown_channel, # message's channel deleted
                        ERROR_CODES.max_reactions, # reached reaction 20, some1 is trolling us.
                        ERROR_CODES.invalid_access, # client removed
                        ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                        ERROR_CODES.cannot_message_user, # user has dm-s disallowed
                            ):
                    return self
            
            raise
        
        self.timeouter = Timeouter(self, timeout=timeout)
        client.events.reaction_add.append(message, self)
        client.events.reaction_delete.append(message, self)
        return self
    
    def _render_embed(self):
        """
        Renders the choose menu's embed's description with it's choices of the respective page and it's footer
        with page information.
        
        Returns
        -------
        embed : ``Embed`` (or any compatible)
            The rendered embed.
        """
        selected = self.selected
        choices = self.choices
        index = (selected//10)*10
        end = index+10
        if len(choices) < end:
            end = len(choices)
        
        parts = []
        prefix = self.prefix
        left_length = 195
        if (prefix is not None):
            left_length -= len(prefix)
        
        while True:
            title = choices[index]
            if isinstance(title,tuple):
                if not title:
                    title = ''
                else:
                    title = title[0]
            
            if not isinstance(title,str):
                title = str(title)
            
            if len(title) > left_length:
                space_position = title.rfind(' ', left_length-25, left_length)
                if space_position == -1:
                    space_position = left_length-3
                
                title = title[:space_position]+'...'
            
            if index == selected:
                if (prefix is not None):
                    parts.append('**')
                    parts.append(prefix)
                    parts.append('** ')
                parts.append('**')
                parts.append(title)
                parts.append('**\n')
            else:
                if (prefix is not None):
                    parts.append(prefix)
                    parts.append(' ')
                parts.append(title)
                parts.append('\n')
            
            index +=1
            if index == end:
                break
        
        embed = self.embed
        embed.description = ''.join(parts)
        
        current_page = (selected//10)+1
        limit = len(choices)
        page_limit = (limit//10)+1
        start = end-9
        if start < 1:
            start = 1
        if end == len(choices):
            end -= 1
        limit -= 1
        
        embed.add_footer(f'Page {current_page}/{page_limit}, {start} - {end} / {limit}, selected: {selected+1}')
        return embed
    
    async def __call__(self, client, event):
        """
        Called when a reaction is added or removed from the respective message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who executes the ``ChooseMenu``
        event : ``ReactionAddEvent``, ``ReactionDeleteEvent``
            The received event.
        """
        if event.user.is_bot:
            return
        
        if (event.emoji not in (self.EMOJIS if len(self.choices)>10 else self.EMOJIS_RESTRICTED)):
            return
        
        client = self.client
        if (event.delete_reaction_with(client) == event.DELETE_REACTION_NOT_ADDED):
            return
        
        check = self.check
        if (check is not None):
            try:
                should_continue = check(event)
            except BaseException as err:
                await client.events.error(client, f'{self!r}.__call__', err)
                return
            
            if not should_continue:
                return
        
        task_flag = self.task_flag
        if task_flag != GUI_STATE_READY:
            if task_flag == GUI_STATE_SWITCHING_PAGE:
                if event.emoji is self.CANCEL:
                    self.task_flag = GUI_STATE_CANCELLING
                return
            
            # ignore GUI_STATE_CANCELLED and GUI_STATE_SWITCHING_CTX
            return
        
        message = self.message
        
        while True:
            emoji = event.emoji
            if emoji is self.UP:
                selected = self.selected-1
                break
            
            if emoji is self.DOWN:
                selected = self.selected+1
                break
            
            if emoji is self.LEFT:
                selected = self.selected-10
                break
            
            if emoji is self.RIGHT:
                selected = self.selected+10
                break
            
            if emoji is self.CANCEL:
                self.task_flag = GUI_STATE_CANCELLED
                try:
                    await client.message_delete(message)
                except BaseException as err:
                    self.cancel()
                    
                    if isinstance(err, ConnectionError):
                        # no internet
                        return
                    
                    if isinstance(err, DiscordException):
                        if err.code in (
                                ERROR_CODES.unknown_channel, # message's channel deleted
                                ERROR_CODES.unknown_message, # message deleted
                                ERROR_CODES.invalid_access, # client removed
                                    ):
                            return
                    
                    await client.events.error(client, f'{self!r}.__call__', err)
                    return
                
                else:
                    self.cancel()
                    return
            
            if emoji is self.SELECT:
                self.task_flag = GUI_STATE_SWITCHING_CTX
                self.cancel()
                
                try:
                    if self.channel.cached_permissions_for(client).can_manage_messages:
                        await client.reaction_clear(message)
                    
                    else:
                        for emoji in self.EMOJIS:
                            await client.reaction_delete_own(message, emoji)
                except BaseException as err:
                    if isinstance(err, ConnectionError):
                        # no internet
                        return
                    
                    if isinstance(err, DiscordException):
                        if err.code in (
                                ERROR_CODES.unknown_message, # message already deleted
                                ERROR_CODES.unknown_channel, # channel deleted
                                ERROR_CODES.invalid_access, # client removed
                                ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                                    ):
                            return
                    
                    await client.events.error(client, f'{self!r}.__call__', err)
                    return
                
                selector = self.selector
                try:
                    choice = self.choices[self.selected]
                    channel = self.channel
                    if isinstance(choice, tuple):
                        coro = selector(client, channel, message, *choice)
                    else:
                        coro = selector(client, channel, message, choice)
                    await coro
                except BaseException as err:
                    await client.events.error(client, f'{self!r}.__call__ when calling {selector!r}', err)
                return
            
            return
        
        if selected < 0:
            selected = 0
        elif selected >= len(self.choices):
            selected = len(self.choices)-1
        
        if self.selected == selected:
            return
        
        self.selected = selected
        self.task_flag = GUI_STATE_SWITCHING_PAGE
        try:
            await client.message_edit(message, embed=self._render_embed())
        except BaseException as err:
            self.task_flag = GUI_STATE_CANCELLED
            self.cancel()
            
            if isinstance(err, ConnectionError):
                # no internet
                return
            
            if isinstance(err, DiscordException):
                if err.code in (
                        ERROR_CODES.unknown_message, # message already deleted
                        ERROR_CODES.unknown_channel, # message's channel deleted
                        ERROR_CODES.invalid_access, # client removed
                            ):
                    return
            
            # We definitely do not want to silence `ERROR_CODES.invalid_form_body`
            await client.events.error(client, f'{self!r}.__call__', err)
            return

        if self.task_flag == GUI_STATE_CANCELLING:
            self.task_flag = GUI_STATE_CANCELLED
            try:
                await client.message_delete(message)
            except BaseException as err:
                
                if isinstance(err, ConnectionError):
                    # no internet
                    return
                
                if isinstance(err, DiscordException):
                    if err.code in (
                            ERROR_CODES.unknown_channel, # channel deleted
                            ERROR_CODES.unknown_message, # message deleted
                            ERROR_CODES.invalid_access, # client removed
                                ):
                        return
                
                await client.events.error(client, f'{self!r}.__call__', err)
                return
            
            self.cancel()
            return
        
        self.task_flag = GUI_STATE_READY
        self.timeouter.set_timeout(self.timeout)
    
    async def _canceller(self, exception,):
        """
        Used when the ``ChooseMenu`` is cancelled.
        
        First of all removes the choose menu from waitfors, so it will not wait for reaction events, then sets the
        ``.task_flag`` of the it to `GUI_STATE_CANCELLED`.
        
        If `exception` is given as `TimeoutError`, then removes the ``ChooseMenu``'s reactions from the respective
        message.
        
        This method is a coroutine.
        
        Parameters
        ----------
        exception : `None` or ``BaseException`` instance
            Exception to cancel the ``ChooseMenu`` with.
        """
        client = self.client
        message = self.message
        
        client.events.reaction_add.remove(message, self)
        client.events.reaction_delete.remove(message, self)
        
        if self.task_flag == GUI_STATE_SWITCHING_CTX:
            # the message is not our, we should not do anything with it.
            return
        
        self.task_flag = GUI_STATE_CANCELLED
        
        if exception is None:
            return
        
        if isinstance(exception, TimeoutError):
            if self.channel.cached_permissions_for(client).can_manage_messages:
                try:
                    await client.reaction_clear(message)
                except BaseException as err:
                    
                    if isinstance(err, ConnectionError):
                        # no internet
                        return
                    
                    if isinstance(err, DiscordException):
                        if err.code in (
                                ERROR_CODES.unknown_message, # message deleted
                                ERROR_CODES.unknown_channel, # channel deleted
                                ERROR_CODES.invalid_access, # client removed
                                ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                                    ):
                            return
                    
                    await client.events.error(client, f'{self!r}._canceller', err)
                    return
            return
        
        timeouter = self.timeouter
        if (timeouter is not None):
            timeouter.cancel()
        # we do nothing
    
    def cancel(self, exception=None):
        """
        Cancels the choose menu, if it is not cancelled yet.
        
        Parameters
        ----------
        exception : `None` or ``BaseException`` instance, Optional
            Exception to cancel the choose menu with. Defaults to `None`
        """
        canceller = self.canceller
        if canceller is None:
            return
        
        self.canceller = None
        
        timeouter = self.timeouter
        if timeouter is not None:
            timeouter.cancel()
        
        return Task(canceller(self, exception), KOKORO)

    def __repr__(self):
        """Returns the choose menu's representation."""
        result = [
            '<', self.__class__.__name__,
            ' client=', repr(self.client),
            ', choices=', repr(len(self.choices)),
            ', selected=', repr(self.selected),
            ', channel=', repr(self.channel),
            ', selector=', repr(self.selector),
                ]
        
        prefix = self.prefix
        if (prefix is not None):
            result.append(', prefix=')
            result.append(repr(prefix))
        
        result.append(', task_flag=')
        task_flag = self.task_flag
        result.append(repr(task_flag))
        result.append(' (')
        
        task_flag_name = (
            'GUI_STATE_READY',
            'GUI_STATE_SWITCHING_PAGE',
            'GUI_STATE_CANCELLING',
            'GUI_STATE_CANCELLED',
            'GUI_STATE_SWITCHING_CTX',
                )[task_flag]
        
        result.append(task_flag_name)
        result.append(')>')
        
        return ''.join(result)

class WaitAndContinue(object):
    """
    Waits for the given event and if the check returns `True` called with the received parameters, then passes them to
    it's waiter future. If check return anything else than `False`, then passes that as well to the future.
    
    Attributes
    -----------
    canceller : `None` `function`
        The canceller function of the ``WaitAndContinue``, what is set to ``._canceller`` by default.
        When ``.cancel`` is called, then this instance attribute is set to `None`.
    check : `callable`
        The check what is called with the received parameters whenever an event is received.
    event : `async-callable`
        The respective event handler on what the waiting is executed.
    future : ``Future``
        The waiter future what's result will be set when the check returns non `False` value.
    target : ``DiscordEntity``
        The target entity on what the waiting is executed.
    timeouter : ``TimeOuter``
        Executes the ``WaitAndContinue`` timeout feature and raise `TimeoutError` to the waiter.
    """
    __slots__ = ('canceller', 'check', 'event', 'future', 'target', 'timeouter', )
    def __init__(self, future, check, target, event, timeout):
        """
        Creates a new ``WaitAndContinue`` instance with the given parameters.
        
        Parameters
        ----------
        future : ``Future`
            The waiter future `what's result will be set when the check returns non `False` value.
        check : `callable`
            The check what is called with the received parameters whenever an event is received.
        target : ``DiscordEntity``
            The target entity on what the waiting is executed.
        event : `async-callable`
            The respective event handler on what the waiting is executed.
        timeout : `float`
            The timeout after `TimeoutError` will be raised to the waiter future.
        """
        self.canceller = self.__class__._canceller
        self.future = future
        self.check = check
        self.event = event
        self.target = target
        self.timeouter = Timeouter(self,timeout)
        event.append(target, self)
    
    async def __call__(self, client, *args):
        """
        Calls the ``WaitAndContinue`` and if it's check returns non `False`, then set's the waiter future's result to
        the received parameters. If `check` returned non `bool`, then passes that value to the waiter as well.
        
        This method is a coroutine.
        
        Parameters
        ----------
        client : ``Client``
            The client who received the respective event.
        *args : `Any`
            Received parameters given by the respective event handler.
        """
        try:
            result = self.check(*args)
        except BaseException as err:
            self.future.set_exception_if_pending(err)
            self.cancel()
        else:
            if type(result) is bool:
                if not result:
                    return
                
                if len(args) == 1:
                    args = args[0]
            
            else:
                args = (*args, result,)
            
            self.future.set_result_if_pending(args)
            self.cancel()
    
    async def _canceller(self, exception):
        """
        Cancels the ``WaitAndContinue`` with the given exception. If the given `exception` is `BaseException` instance,
        then raises it to the waiter future.
        
        This method is a coroutine.
        
        Parameters
        ----------
        exception : `None` or `BaseException`
            Exception to cancel the ``WaitAndContinue``'s ``.future`` with.
        """
        if exception is None:
            self.future.set_exception_if_pending(TimeoutError())
            return
        
        self.event.remove(self.target, self)
        self.future.set_exception_if_pending(exception)
        
        if not isinstance(exception, TimeoutError):
            return
        
        timeouter = self.timeouter
        if (timeouter is not None):
            timeouter.cancel()
    
    def cancel(self):
        """
        Cancels the ``WaitAndContinue``.
        """
        canceller = self.canceller
        if canceller is None:
            return
        
        self.event.remove(self.target, self)
        timeouter = self.timeouter
        if (timeouter is not None):
            timeouter.cancel()
        
        return Task(canceller(self, None), KOKORO)


def wait_for_reaction(client, message, check, timeout):
    """
    Executes waiting for reaction on a message with a ``Future`` instance.
    
    Parameters
    ----------
    client : ``Client``
        The client who's `reaction_add` event will be used.
    message : ``Message``
        The target message on what new reactions will be checked.
    check : `callable`
        The check what is called with the received parameters whenever an event is received.
    timeout : `float`
        The timeout after `TimeoutError` will be raised to the waiter future.
    
    Returns
    -------
    future : ``Future``
        The waiter future, what should be awaited.
    """
    future = Future(KOKORO)
    WaitAndContinue(future, check, message, client.events.reaction_add, timeout)
    return future


def wait_for_message(client, channel, check, timeout):
    """
    Executes waiting for messages at a channel with a ``Future`` instance.
    
    Parameters
    ----------
    client : ``Client``
        The client who's `message_create` event will be used.
    channel : ``ChannelBase``
        The target channel where the new messages will be checked.
    check : `callable`
        The check what is called with the received parameters whenever an event is received.
    timeout : `float`
        The timeout after `TimeoutError` will be raised to the waiter future.
    
    Returns
    -------
    future : ``Future``
        The waiter future, what should be awaited.
    """
    future = Future(KOKORO)
    WaitAndContinue(future, check, channel, client.events.message_create, timeout)
    return future


class _CDUnit(object):
    """
    A cooldown unit stored by a ``CooldownWrapper``.
    
    Attributes
    ----------
    expires_at : `float`
        When the cooldown unit will expire in LOOP_TIME time.
    uses_left : `int`
        How much uses are left till the respective entity will be locked by cooldown.
    """
    __slots__ = ('expires_at', 'uses_left',)
    def __init__(self, expires_at, uses_left):
        """
        Creates a new ``_CDUnit`` with the given parameters.
        
        Parameters
        ----------
        expires_at : `float`
            When the cooldown unit will expire in LOOP_TIME time.
        uses_left : `int`
            How much uses are left till the respective entity will be locked by cooldown.
        """
        self.expires_at = expires_at
        self.uses_left = uses_left
    
    def __repr__(self):
        """Returns the object's representation."""
        return f'{self.__class__.__name__}(expires_at={self.expires_at}, uses_left={self.uses_left})'

class CooldownWrapper(CommandWrapper):
    """
    Rich command wrapper of ``Cooldown``. Check ``CommandWrapper`` itself for more details.
    
    This subclass adds `Cooldown.shared` feature to the command wrappers created by ``Cooldown``.
    """
    __slots__ = ()
    def shared(self, weight=0, func=None):
        """
        Creates a new cooldown instance, which cooldown is shared with the source one.
        
        Parameters
        ----------
        weight : `int`, Optional
            The weight of one call. Defaults to `1`.
        func : `async-callable`, Optional
            The wrapped command. If not given, returns a wrapper, what can be used as a decorator.

        Returns
        -------
        wrapper : ``Cooldown._wrapper`` or ``CooldownWrapper``
            If `func` is given, then returns the created ``CooldownWrapper``, if not, then returns a wrapper,
            what can be used as a decorator.
        
        Raises
        ------
        TypeError
            If `weight` is not numeric convertable to `int`.
        """
        weight_type = weight.__class__
        if weight_type is int:
            pass
        elif issubclass(weight_type, int):
            weight = int(weight)
        else:
            raise TypeError(f'`weight` can be given as `int` instance, got {weight_type.__name__}.') from None
        
        source_wrapper = self.wrapper
        if weight == 0:
            weight = source_wrapper.weight
        
        new_wrapper = object.__new__(type(source_wrapper))
        new_wrapper.checker = source_wrapper.checker
        new_wrapper.reset = source_wrapper.reset
        new_wrapper.cache = source_wrapper.cache
        new_wrapper.weight = weight
        new_wrapper.limit = source_wrapper.limit+source_wrapper.weight-weight
        
        if func is None:
            wrapper = source_wrapper._wrapper(new_wrapper, self.handler)
        else:
            wrapper = CooldownWrapper(func, new_wrapper, self.handler)
        
        return wrapper

class Cooldown(object):
    """
    Helper class for implement cooldowns.
    
    > Rework planned.
    
    Examples
    --------
    
    **Using a cooldown handler example:**
    
    ```
    from hata import DiscordException, CancelledError, sleep, ERROR_CODES, KOKORO
    from hata.ext.commands import Cooldown
    
    class CooldownHandler:
        __slots__ = ('cache',)
        
        def __init__(self):
            self.cache = {}
        
        async def __call__(self, client, message, command, time_left):
            user_id = message.author.id
            try:
                notification,waiter = self.cache[user_id]
            except KeyError:
                pass
            else:
                if notification.channel is message.channel:
                    try:
                        await client.message_edit(notification,
                            f'**{message.author:f}** please cool down, {time_left:.0f} seconds left!')
                    except BaseException as err:
                        if isinstance(err, ConnectionError):
                            return
                        
                        if isinstance(err, DiscordException):
                            if err.code in (
                                    ERROR_CODES.unknown_message, # message deleted
                                    ERROR_CODES.unknown_channel, # channel deleted
                                    ERROR_CODES.invalid_access, # client removed
                                        ):
                                return
                        
                        await client.events.error(client, f'{self!r}.__call__', err)
                    
                    return
                
                waiter.cancel()
            
            try:
                notification = await client.message_create(message.channel,
                    f'**{message.author:f}** please cool down, {time_left:.0f} seconds left!')
            except BaseException as err:
                if isinstance(err, ConnectionError):
                    return
                
                if isinstance(err, DiscordException):
                    if err.code in (
                            ERROR_CODES.unknown_message, # message deleted
                            ERROR_CODES.unknown_channel, # message's channel deleted
                            ERROR_CODES.invalid_access, # client removed
                            ERROR_CODES.invalid_permissions, # permissions changed meanwhile
                            ERROR_CODES.cannot_message_user, # user has dm-s disallowed
                                ):
                        return
                
                await client.events.error(client, f'{self!r}.__call__', err)
            
            waiter = Task(self.waiter(client, user_id, notification), KOKORO)
            self.cache[user_id] = (notification, waiter)
        
        async def waiter(self, client, user_id, notification):
            try:
                await sleep(30., KOKORO)
            except CancelledError:
                pass
            
            del self.cache[user_id]
            
            try:
                await client.message_delete(notification)
            except BaseException as err:
                if isinstance(err, ConnectionError):
                    # no internet
                    return
                
                if isinstance(err, DiscordException):
                    if err.code in (
                            ERROR_CODES.unknown_channel, # message's channel deleted
                            ERROR_CODES.unknown_message, # message deleted
                            ERROR_CODES.invalid_access, # client removed
                                ):
                        return
                
                await client.events.error(client, f'{self!r}.__call__', err)
    
    @Bot.commands
    @Cooldown('user', 30., handler=CooldownHandler())
    async def ping(client, message):
        await client.message_create(message.channel, f'{client.gateway.latency.:.0f} ms')
    ```
    
    **Using shared cooldowns:**

    ```
    from hata import Embed
    from hata.ext.commands import Converter, ConverterFlag
    
    @Bot.commands
    @Cooldown('user', 60., limit=3, weight=2, handler=CooldownHandler())
    async def avatar(client, message, user : Converter('user', flags=ConverterFlag.user_default.update_by_keys(everywhere=True), default_code='message.author')):
        url = user.avatar_url_as(size=4096)
        embed = Embed(f'{user:f}\'s avatar', url=url)
        embed.add_image(url)
        await client.message_create(message.channel, embed=embed)
    
    @Bot.commands
    @avatar.shared(weight=1)
    async def myavatar(client, message):
        url = message.author.avatar_url_as(size=4096)
        embed = Embed('Your avatar', url=url)
        embed.add_image(url)
        await client.message_create(message.channel, embed=embed)
    ```
    
    Attributes
    ----------
    cache : `dict` of (``DiscordEntity``, ``_CDUnit``) items
        Cache to remember how much use of the given entity are exhausted already.
    checker : `function`
        Checks after how much time the given entity can use again the respective command.
    limit : `int`
        The amount of how much times the command can be called within a set duration before going on cooldown.
    reset : `float`
        The time after the cooldown resets.
    weight : `int`
        The weight of the command.
    """
    __slots__ = ('cache', 'checker', 'limit', 'reset', 'weight',)
    
    def __new__(cls, for_, reset, limit=1, weight=1, handler=None, func=None):
        """
        Creates a new
        
        Parameters
        ----------
        for_ : `str`
            By what type of entity the cooldown should limit the command.
            
            Possible values:
             - `'user'`
             - `'channel'`
             - `'guild'`
         
        reset : `float`
            The reset time of the cooldown.
        limit : `int`
            The amount of calls after the respective command goes on cooldown.
        weight : `int`, Optional
            The weight of one call. Defaults to `1`.
        handler : `None` or `async-callable`
            Called, when the wrapped command is on cooldown.
            
            If given then 4 parameters will be passed to it:
            +-------------------+---------------+
            | Respective name   | Type          |
            +===================+===============+
            | client            | ``Client``    |
            +-------------------+---------------+
            | message           | ``Message``   |
            +-------------------+---------------+
            | command           | ``Command``   |
            +-------------------+---------------+
            | time_left         | `float`       |
            +-------------------+---------------+
        
        func : `async-callable`, Optional
            The wrapped command. If not given, returns a wrapper, what can be used as a decorator.
        
        Returns
        -------
        wrapper : ``Cooldown._wrapper`` / ``CooldownWrapper``
            If `func` is given, then returns the created ``CooldownWrapper``, if not, then returns a wrapper,
            what can be used as a decorator.
        
        Raises
        ------
        TypeError
            - If `str` is not given as `str` instance.
            - If `weight` is not numeric convertable to `int`.
            - If `reset` is not numeric convertable to `float`.
            - If `limit` is not numeric convertable to `int`.
        ValueError
            - If `for_` is not given as any of the expected value.
        """
        for_type = for_.__class__
        if for_type is str:
            pass
        elif issubclass(for_, str):
            for_ = str(for_)
        else:
            raise TypeError(f'`for_` can be given as `str` instance, got {for_type.__name__}.')
    
        if 'user'.startswith(for_):
            checker = cls._check_user
        elif 'channel'.startswith(for_):
            checker = cls._check_channel
        elif 'guild'.startswith(for_):
            checker = cls._check_guild
        else:
            raise ValueError(f'\'for_\' can be \'user\', \'channel\' or \'guild\', got {for_!r}')
        
        reset_type = reset.__class__
        if (reset_type is not float):
            try:
                __float__ = getattr(reset_type, '__float__')
            except AttributeError:
                raise TypeError(f'The given reset is not `float`, neither other numeric convertable to it, got '
                    f'{reset_type.__name__}.') from None
            
            reset = __float__(reset)
            
        limit_type = limit.__class__
        if limit_type is int:
            pass
        elif issubclass(limit_type, int):
            limit = int(limit)
        else:
            raise TypeError(f'`limit` can be given as `int` instance, got {limit_type.__name__}.') from None
        
        weight_type = weight.__class__
        if weight_type is int:
            pass
        elif issubclass(weight_type, int):
            weight = int(weight)
        else:
            raise TypeError(f'`weight` can be given as `int` instance, got {weight_type.__name__}.') from None
        
        self = object.__new__(cls)
        self.checker = checker
        self.reset = reset
        self.weight = weight
        self.limit = limit-weight
        self.cache = {}
        
        if func is None:
            wrapper = self._wrapper(self, handler)
        else:
            wrapper = CooldownWrapper(func, self, handler)
        
        return wrapper
    
    class _wrapper(object):
        """
        When a parent ``Command`` instance would be created without giving `func` parameter, then a wrapper of this
        type is returned enabling using ``Cooldown` as a decorator, with still giving parameters to it.
        
        Attributes
        ----------
        parent : ``Cooldown``
            The parent cooldown instance.
        handler : `None` or `async-callable`
            Called, when the wrapped command is on cooldown.
            
            If given then 4 parameters will be passed to it:
            +-------------------+---------------+
            | Respective name   | Type          |
            +===================+===============+
            | client            | ``Client``    |
            +-------------------+---------------+
            | message           | ``Message``   |
            +-------------------+---------------+
            | command           | ``Command``   |
            +-------------------+---------------+
            | time_left         | `float`       |
            +-------------------+---------------+
        """
        __slots__ = ('parent', 'handler')
        def __init__(self, parent, handler):
            """
            Creates a new ``Cooldown._wrapper`` instance with the given parameters.
            
            Parameters
            ----------
            parent : ``Cooldown``
                The parent cooldown instance.
            handler : `None` or `async-callable`
                Called, when the wrapped command is on cooldown.
                
                If given then 4 parameters will be passed to it:
                +-------------------+---------------+
                | Respective name   | Type          |
                +===================+===============+
                | client            | ``Client``    |
                +-------------------+---------------+
                | message           | ``Message``   |
                +-------------------+---------------+
                | command           | ``Command``   |
                +-------------------+---------------+
                | time_left         | `float`       |
                +-------------------+---------------+
            """
            self.parent = parent
            self.handler = handler
        
        def __call__(self, func):
            """
            By calling a cooldown's ``._wrapper`` a ``CooldownWrapper`` instance is created and returned, what can be
            added as a command.
            
            Parameters
            ----------
            func : `async-callable`
                The wrapped function by the cooldown to add as a command.

            Returns
            -------
            wrapper : ``CooldownWrapper``
            """
            if func is None:
                raise TypeError('`func` is given as `None`.')
            
            return CooldownWrapper(func, self.parent, self.handler)
    
    async def __call__(self, client, message):
        """
        Calls the cooldown with the respective `client` and `message`, and then yields whether the command can be
        called, and if not, then with what extra parameters the handler should receive.
        
        This method is a coroutine generator.
        
        Parameters
        ----------
        client : ``Client``
            The client who received the respective message.
        message : ``Message``
            The received message.
        
        Yields
        ------
        passed : `bool`
            Whether the command can be called. If not, then yields additional parameters to call the cooldown's
            handler with.
        time_left : `float`
            How much time is left till the cooldown's expiration.
        """
        value = self.checker(self, message)
        if not value:
            yield True
            return
        
        yield False
        yield value-LOOP_TIME()
        return
    
    @staticmethod
    def _check_user(self, message):
        """
        Executes user cooldown check.
        
        Might be set as the ``Cooldown``'s ``.checker`` instance attribute.
        
        Parameters
        ----------
        message : ``Message``
            The received message.
        
        Returns
        -------
        expires_at : `int`
            When the cooldown for the given entity will expire.
        """
        id_ = message.author.id
        
        cache = self.cache
        try:
            unit = cache[id_]
        except KeyError:
            at_ = LOOP_TIME()+self.reset
            cache[id_] = _CDUnit(at_, self.limit)
            KOKORO.call_at(at_, dict.__delitem__, cache, id_)
            return 0.
        
        left = unit.uses_left
        if left > 0:
            unit.uses_left = left-self.weight
            return 0.
        return unit.expires_at
    
    @staticmethod
    def _check_channel(self, message):
        """
        Executes channel cooldown check.
        
        Might be set as the ``Cooldown``'s ``.checker`` instance attribute.
        
        Parameters
        ----------
        message : ``Message``
            The received message.
        
        Returns
        -------
        expires_at : `int`
            When the cooldown for the given entity will expire.
        """
        id_ = message.channel.id
        
        cache = self.cache
        try:
            unit = cache[id_]
        except KeyError:
            at_ = LOOP_TIME()+self.reset
            cache[id_] = _CDUnit(at_, self.limit)
            KOKORO.call_at(at_, dict.__delitem__, cache, id_)
            return 0.
        
        left = unit.uses_left
        if left>0:
            unit.uses_left = left-self.weight
            return 0.
        return unit.expires_at
    
    #returns -1. if non guild
    @staticmethod
    def _check_guild(self, message):
        """
        Executes guild based cooldown check.
        
        Might be set as the ``Cooldown``'s ``.checker`` instance attribute.
        
        Parameters
        ----------
        message : ``Message``
            The received message.
        
        Returns
        -------
        expires_at : `int`
            When the cooldown for the given entity will expire.
            
            If the cooldown limitation is not applicable for the given entity, returns `-1.0`.
        """
        channel = message.channel
        if channel.type in (1, 3):
            return -1.
        else:
            id_ = channel.guild.id
        
        cache = self.cache
        try:
            unit = cache[id_]
        except KeyError:
            at_ = LOOP_TIME()+self.reset
            cache[id_] = _CDUnit(at_,self.limit)
            KOKORO.call_at(at_, dict.__delitem__, cache, id_)
            return 0.
        
        left = unit.uses_left
        if left > 0:
            unit.uses_left = left-self.weight
            return 0.
        return unit.expires_at



# -*- coding: utf-8 -*-
from ..extension_loader.snapshot import SNAPSHOT_TAKERS
from .command import Slasher

def take_slasher_snapshot(client):
    """
    Collects all the command changes from the client's slash command processor.
    
    Parameters
    ----------
    client : ``Client``
        The client, who will be snapshotted.
    
    Returns
    -------
    collected : `None` or `set` of ``SlashCommand``
    """
    slasher = getattr(client, 'slasher', None)
    if (slasher is None) or (not isinstance(slasher, Slasher)):
        collected = None
    
    else:
        collected = slasher._get_commands()
    
    return collected

def calculate_slasher_snapshot_difference(client, snapshot_old, snapshot_new):
    """
    Calculates the difference between two slasher snapshots
    
    Parameters
    ----------
    client : ``Client``
        The respective client.
    snapshot_old : `None` of `set` of ``SlashCommand``
        An old snapshot taken.
    snapshot_new : `None` of `set` of ``SlashCommand``
        A new snapshot.
    
    Returns
    -------
    snapshot_difference : `None` or `tuple` (`set` of ``SlashCommand``, `set` of ``SlashCommand``)
    """
    if (snapshot_old is None) or (snapshot_new is None):
        return None
    
    removed_commands = []
    
    while snapshot_old:
        old_command = snapshot_old.pop()
        
        try:
            snapshot_new.remove(snapshot_new)
        except KeyError:
            removed_commands.append(old_command)
    
    added_commands = snapshot_new-snapshot_old
    removed_commands = snapshot_old-snapshot_new
    
    if (not added_commands) and (not removed_commands):
        return None
    
    if client.running and client.application.id:
        slasher = getattr(client, 'slasher', None)
        if (slasher is not None):
            slasher.do_main_sync(client)
    
    return added_commands, removed_commands

def revert_slasher_snapshot(client, snapshot_difference):
    """
    Reverts a snapshot taken from a slasher.
    
    Parameters
    ----------
    client : ``Client``
        The respective client instance.
    snapshot_difference : `tuple` (`set` of ``SlashCommand``, `set` of ``SlashCommand``)
        The taken snapshot.
    """
    slasher = getattr(client, 'slasher', None)
    if (slasher is None) or (not isinstance(slasher, Slasher)):
        return
    
    added_commands, removed_commands = snapshot_difference
    for command in added_commands:
        slasher._remove_command(command)
    
    for command in removed_commands:
        slasher._add_command(command)
    
    slasher.do_main_sync(client)

SNAPSHOT_TAKERS['client.slasher'] = (
    take_slasher_snapshot,
    calculate_slasher_snapshot_difference,
    revert_slasher_snapshot,
        )



# -*- coding: utf-8 -*-
"""
Hata extension for controlling slash commands.

Work in progress.
"""
from ...backend.futures import Task
from ...discord.client_core import KOKORO

from .command import Slasher
from .utils import _do_initial_sync, _application_command_create_watcher, _application_command_delete_watcher, \
    delay_immediate_start_initial_sync
from .client_wrapper_extension import *

__all__ = command.__all__

def setup_ext_slash(client, *, immediate_sync=False):
    """
    Setups the slash extension on client.
    
    Note, that this function can be called on a client only once.
    
    Parameters
    ----------
    client : ``Client``
        The client to setup the extension on.
    immediate_sync : `bool`, Optional
        Whether application command sync task should be started immediately as it is detected that no more features
        are added to the slasher. Defaults to `False`.
    
    Returns
    -------
    slasher : ``Slasher``
        Slash command processor.
    
    Raises
    ------
    RuntimeError
        If the client has an attribute set what the slasher would use.
    """
    for attr_name in ('slasher', 'interactions'):
        if hasattr(client, attr_name):
            raise RuntimeError(f'The client already has an attribute named as `{attr_name}`.')
    
    slasher = Slasher()
    
    client.events(slasher)
    client.slasher = slasher
    client.interactions = slasher.shortcut
    client.events(_do_initial_sync, name='ready')
    client.events(_application_command_create_watcher, name='application_command_create')
    client.events(_application_command_delete_watcher, name='application_command_delete')
    
    if immediate_sync:
        Task(delay_immediate_start_initial_sync(client, slasher), KOKORO)
    
    return slasher


def snapshot_hook():
    from . import snapshot
    

from .. import register_library_extension, add_library_extension_hook
register_library_extension('HuyaneMatsu.slash')
add_library_extension_hook(snapshot_hook, ['HuyaneMatsu.extension_loader'])
del register_library_extension, add_library_extension_hook, snapshot_hook
